#!/usr/bin/env python3
#
# Copyright 2023 Richard Hughes <richard@hughsie.com>
#
# SPDX-License-Identifier: LGPL-2.1-or-later
#
# pylint: disable=invalid-name,missing-docstring,consider-using-f-string

import os
import datetime
import argparse
import glob
import sys
from jinja2 import Environment, FileSystemLoader, select_autoescape

subst = {}


# convert a snake-case name into CamelCase
def _to_camelcase(value: str) -> str:
    return "".join([tmp[0].upper() + tmp[1:].lower() for tmp in value.split("_")])


def _subst_replace(data: str) -> str:
    for key, value in subst.items():
        data = data.replace(key, value)
    return data


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--vendor",
        type=str,
        help="Vendor name",
        required=True,
    )
    parser.add_argument(
        "--example",
        type=str,
        help="Plugin basename",
        required=True,
    )
    parser.add_argument("--parent", type=str, default="Usb", help="Device parent GType")
    parser.add_argument(
        "--year", type=int, default=datetime.date.today().year, help="Copyright year"
    )
    parser.add_argument("--author", type=str, help="Copyright author", required=True)
    parser.add_argument("--email", type=str, help="Copyright email", required=True)
    args = parser.parse_args()

    try:
        vendor: str = args.vendor.replace("-", "_")
        example: str = args.example.replace("-", "_")

        # first in list
        subst["VendorExample"] = _to_camelcase(vendor) + _to_camelcase(example)
        subst["vendor_example"] = vendor.lower() + "_" + example.lower()
        subst["vendor_dash_example"] = vendor.lower() + "-" + example.lower()
        subst["VENDOR_EXAMPLE"] = vendor.upper() + "_" + example.upper()

        # second
        for key, value in {
            "Vendor": vendor,
            "Example": example,
            "Parent": args.parent,
            "Year": str(args.year),
            "Author": args.author,
            "Email": args.email,
        }.items():
            subst[key] = _to_camelcase(value)
            subst[key.lower()] = value.lower()
            subst[key.upper()] = value.upper()

    except NotImplementedError as e:
        print(e)
        sys.exit(1)

    template_src: str = "vendor-example"
    os.makedirs(os.path.join("plugins", _subst_replace(template_src)), exist_ok=True)

    srcdir: str = sys.argv[0].rsplit("/", maxsplit=2)[0]
    env = Environment(
        loader=FileSystemLoader(srcdir),
        autoescape=select_autoescape(),
        keep_trailing_newline=True,
    )
    for fn in glob.iglob(f"{srcdir}/plugins/{template_src}/**", recursive=True):
        if os.path.isdir(fn):
            fn_rel: str = os.path.relpath(fn, srcdir)
            os.makedirs(_subst_replace(fn_rel), exist_ok=True)
            continue
        fn_rel: str = os.path.relpath(fn, srcdir)
        template = env.get_template(fn_rel)
        filename: str = _subst_replace(fn_rel.replace(".in", ""))
        with open(filename, "wb") as f_dst:
            f_dst.write(template.render(subst).encode())
        print(f"wrote {filename}")



#!/usr/bin/env python3
# pylint: disable=invalid-name,missing-module-docstring,missing-function-docstring
#
# Copyright 2023 Richard Hughes <richard@hughsie.com>
#
# SPDX-License-Identifier: LGPL-2.1-or-later

import glob
import sys
import os
from typing import List, Optional


def _find_func_name(line: str) -> Optional[str]:
    # these are not functions
    for prefix in ["\t", " ", "typedef", "__attribute__", "G_DEFINE_"]:
        if line.startswith(prefix):
            return None

    # ignore prototypes
    if line.endswith(";"):
        return None

    # strip to function name then test for validity
    idx: int = line.find("(")
    if idx == -1:
        return None
    func_name = line[:idx]
    if func_name.find(" ") != -1:
        return None

    # success!
    return func_name


class SourceFailure:
    def __init__(self, fn=None, linecnt=None, message=None, nocheck=None):
        self.fn: Optional[str] = fn
        self.linecnt: Optional[int] = linecnt
        self.message: Optional[str] = message
        self.nocheck: Optional[str] = nocheck


class Checker:
    MAX_FUNCTION_LINES: int = 400

    def __init__(self):
        self.failures: List[SourceFailure] = []
        self._current_fn: Optional[str] = None
        self._current_linecnt: Optional[int] = None
        self._current_nocheck: Optional[str] = None

    def add_failure(self, message=None):
        self.failures.append(
            SourceFailure(
                fn=self._current_fn,
                linecnt=self._current_linecnt,
                message=message,
                nocheck=self._current_nocheck,
            )
        )

    def _test_line_function_names_private(self, func_name: str) -> None:
        valid_prefixes = ["_fwupd_", "_fu_", "_g_", "_xb_"]
        for prefix in valid_prefixes:
            if func_name.startswith(prefix):
                return
        self.add_failure(
            f"invalid function name {func_name} should have {'|'.join(valid_prefixes)} prefix"
        )

    def _test_line_function_names_valid(self, func_name: str) -> None:
        # sanity check
        if not self._current_fn:
            return

        # ignore headers
        if self._current_fn.endswith(".h"):
            return

        # ignore all self tests FIXME: 335 failures
        if self._current_fn.endswith("-test.c"):
            return

        # namespacing is strange here
        if self._current_fn.endswith("fwupd-enums.c"):
            return

        # doh
        if func_name in ["main", "fu_plugin_init_vfuncs"]:
            return

        # this is stuff that should move to GLib
        if func_name.startswith("_g_"):
            return

        # remove some suffixes we do not care about
        prefix = os.path.basename(self._current_fn).split(".")[0].replace("-", "_")
        for suffix in [
            "_common",
            "_darwin",
            "_freebsd",
            "_helper",
            "_impl",
            "_linux",
            "_sync",
            "_windows",
        ]:
            if prefix.endswith(suffix):
                prefix = prefix[: -len(suffix)]

        # allowed truncations
        valid_prefixes = []
        valid_prefixes.append(prefix)
        for key, value in {
            "fu_crc": "fu_misr",  # FIXME: split out to fu-misr.[c|h]
            "fu_darwin_efivars": "fu_efivars",
            "fu_dbus_daemon": "fu_daemon",
            "fu_dbxtool": "fu_util",
            "fu_freebsd_efivars": "fu_efivars",
            "fu_linux_efivars": "fu_efivars",
            "fu_logitech_hidpp_hidpp": "fu_logitech_hidpp",
            "fu_logitech_hidpp_hidpp_msg": "fu_logitech_hidpp",
            "fu_self_test": "fu_test",
            "fu_string": "fu_str,fu_utf",
            "fu_tool": "fu_util",
            "fu_windows_efivars": "fu_efivars",
        }.items():
            if prefix == key:
                valid_prefixes.extend(value.split(","))
        for prefix in valid_prefixes:
            if func_name.startswith(prefix):
                return

        self.add_failure(
            f"invalid function name {func_name} should have {'|'.join(valid_prefixes)} prefix"
        )

    def _test_line_function_names(self, line: str) -> None:
        # empty line
        if not line:
            return

        # skip!
        self._current_nocheck = "nocheck:name"
        if line.find(self._current_nocheck) != -1:
            return

        # parse
        func_name: str = _find_func_name(line)
        if not func_name:
            return
        if func_name.startswith("_"):
            self._test_line_function_names_private(func_name)
            return
        self._test_line_function_names_valid(func_name)

    def _test_line_enums(self, line: str) -> None:
        # skip!
        self._current_nocheck = "nocheck:prefix"
        if line.find(self._current_nocheck) != -1:
            return

        # needs Fu prefix
        enum_name = None
        valid_prefixes = ["Fwupd", "Fu"]
        if line.startswith("enum ") and line.endswith("{"):
            enum_name = line[5:-2]
        elif line.startswith("typedef enum ") and line.endswith("{"):
            enum_name = line[13:-2]
        if not enum_name:
            return
        for prefix in valid_prefixes:
            if enum_name.startswith(prefix):
                return
        self.add_failure(
            f"invalid enum name {enum_name} should have {'|'.join(valid_prefixes)} prefix"
        )

    def _test_line_debug_fns(self, line: str) -> None:
        # no console output expected
        self._current_nocheck = "nocheck:print"
        if line.find(self._current_nocheck) != -1:
            return
        if self._current_fn and os.path.basename(self._current_fn) in [
            "fu-console.c",
            "fu-daemon.c",
            "fu-dbxtool.c",
            "fu-debug.c",
            "fu-fuzzer-main.c",
            "fu-gcab.c",
            "fu-main.c",
            "fu-main-windows.c",
            "fu-self-test.c",
            "fu-tpm-eventlog.c",
            "fwupd-self-test.c",
        ]:
            return
        for token, msg in {
            "g_print(": "Use g_debug() instead",
            "g_printerr(": "Use g_debug() instead",
        }.items():
            if line.find(token) != -1:
                self.add_failure(f"contains blocked token {token}: {msg}")

    def _test_line_blocked_fns(self, line: str) -> None:
        self._current_nocheck = "nocheck:blocked"
        if line.find(self._current_nocheck) != -1:
            return
        for token, msg in {
            "cbor_get_uint8(": "Use cbor_get_int() instead",
            "cbor_get_uint16(": "Use cbor_get_int() instead",
            "cbor_get_uint32(": "Use cbor_get_int() instead",
            "g_error(": "Use GError instead",
            "g_byte_array_free_to_bytes(": "Use g_bytes_new() instead",
            "g_usb_device_bulk_transfer(": "Use fu_usb_device_bulk_transfer() instead",
            "g_usb_device_claim_interface(": "Use fu_usb_device_claim_interface() instead",
            "g_usb_device_control_transfer(": "Use fu_usb_device_control_transfer() instead",
            "g_usb_device_get_configuration_index(": "Use fu_usb_device_get_configuration_index() instead",
            "g_usb_device_get_custom_index(": "Use fu_usb_device_get_custom_index() instead",
            "g_usb_device_get_device_class(": "Use fu_usb_device_get_class() instead",
            "g_usb_device_get_interface(": "Use fu_usb_device_get_interface() instead",
            "g_usb_device_get_interfaces(": "Use fu_usb_device_get_interfaces() instead",
            "g_usb_device_get_release(": "Use fu_usb_device_get_release() instead",
            "g_usb_device_get_serial_number_index(": "Use fu_usb_device_get_serial_number_index() instead",
            "g_usb_device_get_string_descriptor_bytes_full(": "Use fu_usb_device_get_string_descriptor_bytes_full() instead",
            "g_usb_device_get_string_descriptor desc_index(": "Use fu_usb_device_get_string_descriptor desc_index() instead",
            "g_usb_device_interrupt_transfer(": "Use fu_usb_device_interrupt_transfer() instead",
            "g_usb_device_release_interface(": "Use fu_usb_device_release_interface() instead",
            "g_usb_device_reset error(": "Use fu_usb_device_reset error() instead",
            "g_usb_device_set_interface_alt(": "Use fu_usb_device_set_interface_alt() instead",
            "g_ascii_strtoull(": "Use fu_strtoull() instead",
            "g_ascii_strtoll(": "Use fu_strtoll() instead",
            "g_random_int_range(": "Use a predicatable token instead",
            "g_assert(": "Use g_set_error() or g_return_val_if_fail() instead",
            "g_udev_device_get_sysfs_attr(": "Use fu_udev_device_read_sysfs() instead",
            "g_udev_device_get_property(": "Use fu_udev_device_read_property() instead",
            "g_udev_client_new(": "Use fu_backend_create_device() instead",
            "HIDIOCSFEATURE": "Use fu_hidraw_device_set_feature() instead",
            "HIDIOCGFEATURE": "Use fu_hidraw_device_get_feature() instead",
            "|= 1 <<": "Use FU_BIT_SET() instead",
            "|= 1u <<": "Use FU_BIT_SET() instead",
            "|= 1ull <<": "Use FU_BIT_SET() instead",
            "|= (1 <<": "Use FU_BIT_SET() instead",
            "|= (1u <<": "Use FU_BIT_SET() instead",
            "|= (1ull <<": "Use FU_BIT_SET() instead",
            "&= ~(1 <<": "Use FU_BIT_CLEAR() instead",
            "&= ~(1u <<": "Use FU_BIT_CLEAR() instead",
            "&= ~(1ull <<": "Use FU_BIT_CLEAR() instead",
            "__attribute__((packed))": "Use rustgen instead",
            "memcpy(": "Use fu_memcpy_safe or rustgen instead",
            "GUINT16_FROM_BE(": "Use fu_memread_uint16_safe() or rustgen instead",
            "GUINT16_FROM_LE(": "Use fu_memread_uint16_safe() or rustgen instead",
            "GUINT16_TO_BE(": "Use fu_memwrite_uint16_safe() or rustgen instead",
            "GUINT16_TO_LE(": "Use fu_memwrite_uint16_safe() or rustgen instead",
            "GUINT32_FROM_BE(": "Use fu_memread_uint32_safe() or rustgen instead",
            "GUINT32_FROM_LE(": "Use fu_memread_uint32_safe() or rustgen instead",
            "GUINT32_TO_BE(": "Use fu_memwrite_uint32_safe() or rustgen instead",
            "GUINT32_TO_LE(": "Use fu_memwrite_uint32_safe() or rustgen instead",
            "GUINT64_FROM_BE(": "Use fu_memread_uint64_safe() or rustgen instead",
            "GUINT64_FROM_LE(": "Use fu_memread_uint64_safe() or rustgen instead",
            "GUINT64_TO_BE(": "Use fu_memwrite_uint64_safe() or rustgen instead",
            "GUINT64_TO_LE(": "Use fu_memwrite_uint64_safe() or rustgen instead",
            " ioctl(": "Use fu_udev_device_ioctl() instead",
        }.items():
            if line.find(token) != -1:
                self.add_failure(f"contains blocked token {token}: {msg}")

    def _test_lines_gerror(self, lines: List[str]) -> None:
        self._current_nocheck = "nocheck:error"
        linecnt_g_set_error: int = 0
        for linecnt, line in enumerate(lines):
            if line.find(self._current_nocheck) != -1:
                continue
            self._current_linecnt = linecnt + 1

            # do not use G_IO_ERROR internally
            if line.find("g_set_error") != -1:
                linecnt_g_set_error = linecnt
            if linecnt - linecnt_g_set_error < 5:
                for error_domain in ["G_IO_ERROR", "G_FILE_ERROR"]:
                    if line.find(error_domain) != -1:
                        self.add_failure("uses g_set_error() without using FWUPD_ERROR")
                        break

    def _test_lines_function_length(self, lines: List[str]) -> None:
        self._current_nocheck = "nocheck:lines"
        func_begin: int = 0
        func_name: Optional[str] = None
        for linecnt, line in enumerate(lines):
            if line.find(self._current_nocheck) != -1:
                func_begin = 0
                continue
            if line == "{":
                func_begin = linecnt
                continue
            if func_begin > 0 and line == "}":
                self._current_linecnt = func_begin
                if linecnt - func_begin > self.MAX_FUNCTION_LINES:
                    if func_name:
                        self.add_failure(
                            f"{func_name} is too long, was {linecnt - func_begin} of {self.MAX_FUNCTION_LINES}"
                        )
                    else:
                        self.add_failure(
                            f"function is too long, was {linecnt - func_begin} of {self.MAX_FUNCTION_LINES}"
                        )
                if func_name and linecnt - func_begin < 3:
                    if func_name.endswith("_finalize"):
                        self.add_failure(f"{func_name} is redundant and can be removed")
                func_begin = 0
                func_name = None
                continue

            # is a function?
            func_name_tmp: Optional[str] = _find_func_name(line)
            if func_name_tmp:
                func_name = func_name_tmp

    def _test_lines_firmware_convert_version(self, lines: List[str]) -> None:
        self._current_nocheck = "nocheck:set-version"

        if self._current_fn and os.path.basename(self._current_fn) in [
            "fu-firmware.c",
            "fu-firmware.h",
        ]:
            return

        # contains fu_firmware_set_version_raw()
        _set_version_raw: bool = False
        for line in lines:
            if line.find("fu_firmware_set_version_raw(") != -1:
                _set_version_raw = True
                break
        if not _set_version_raw:
            return

        # also contains fu_firmware_set_version()
        for linecnt, line in enumerate(lines):
            if line.find(self._current_nocheck) != -1:
                continue
            self._current_linecnt = linecnt + 1
            if line.find("fu_firmware_set_version(") != -1:
                self.add_failure(
                    "Use FuFirmwareClass->convert_version rather than fu_firmware_set_version()"
                )

    def _test_lines_device_convert_version(self, lines: List[str]) -> None:
        self._current_nocheck = "nocheck:set-version"

        if self._current_fn and os.path.basename(self._current_fn) in [
            "fu-device.c",
            "fu-device.h",
            "fu-self-test.c",
        ]:
            return

        # contains fu_firmware_set_version_raw()
        _set_version_raw: bool = False
        for line in lines:
            if line.find("fu_device_set_version_raw(") != -1:
                _set_version_raw = True
                break
        if not _set_version_raw:
            return

        # also contains fu_firmware_set_version()
        for linecnt, line in enumerate(lines):
            if line.find(self._current_nocheck) != -1:
                continue
            self._current_linecnt = linecnt + 1
            if line.find("fu_device_set_version(") != -1:
                self.add_failure(
                    "Use FuDeviceClass->convert_version rather than fu_device_set_version()"
                )

    def _test_lines_depth(self, lines: List[str]) -> None:
        # check depth
        self._current_nocheck = "nocheck:depth"
        depth: int = 0
        for linecnt, line in enumerate(lines):
            if line.find(self._current_nocheck) != -1:
                continue
            self._current_linecnt = linecnt + 1
            for char in line:
                if char == "{":
                    depth += 1
                    if depth > 5:
                        self.add_failure("is nested too deep")
                        success = False
                    continue
                if char == "}":
                    if depth == 0:
                        self.add_failure("has unequal nesting")
                        success = False
                        continue
                    depth -= 1

        # sanity check
        self._current_linecnt = None
        if depth != 0:
            self.add_failure("nesting was weird")
            success = False

    def _test_lines(self, lines: List[str]) -> None:
        lines_nocheck: List[str] = []

        # tests we can do line by line
        for linecnt, line in enumerate(lines):
            self._current_linecnt = linecnt + 1

            # test for blocked functions
            self._test_line_blocked_fns(line)

            # test for debug lines
            self._test_line_debug_fns(line)

            # test for function names
            self._test_line_function_names(line)

            # test for invalid enum names
            self._test_line_enums(line)

        # using FUWPD_ERROR domains
        self._test_lines_gerror(lines)

        # not nesting too deep
        self._test_lines_depth(lines)

        # functions too long
        self._test_lines_function_length(lines)

        # should use FuFirmwareClass->convert_version
        self._test_lines_firmware_convert_version(lines)

        # should use FuDeviceClass->convert_version
        self._test_lines_device_convert_version(lines)

    def test_file(self, fn: str) -> None:
        self._current_fn = fn
        with open(fn, "rb") as f:
            try:
                self._test_lines(f.read().decode().split("\n"))
            except UnicodeDecodeError as e:
                print(f"failed to read {fn}: {e}")


def test_files() -> int:
    # test all C and H files
    rc: int = 0

    checker = Checker()

    # use any file specified in argv, falling back to scanning the entire tree
    fns: List[str] = []
    if len(sys.argv) > 1:
        for fn in sys.argv[1:]:
            try:
                ext: str = fn.rsplit(".", maxsplit=1)[1]
            except IndexError:
                continue
            if ext in ["c", "h"]:
                fns.append(fn)
    else:
        fns.extend(glob.glob("libfwupd/*.[c|h]"))
        fns.extend(glob.glob("libfwupdplugin/*.[c|h]"))
        fns.extend(glob.glob("plugins/*/*.[c|h]"))
        fns.extend(glob.glob("src/*.[c|h]"))
    for fn in fns:
        if os.path.basename(fn) == "check-source.py":
            continue
        checker.test_file(fn)

    # show issues
    for failure in checker.failures:
        line: str = ""
        if failure.fn:
            line += failure.fn
        if failure.linecnt:
            line += f":{failure.linecnt}"
        line += f": {failure.message}"
        if failure.nocheck:
            line += f" -- use a {failure.nocheck} comment to ignore"
        print(line)
    if checker.failures:
        print(f"{len(checker.failures)} failures!")
    return 1 if checker.failures else 0


if __name__ == "__main__":
    # all done!
    sys.exit(test_files())



#!/usr/bin/env python3
#
# Copyright 2024 Maciej Borzecki <maciej.borzecki@canonical.cm>
#
# SPDX-License-Identifier: LGPL-2.1-or-later

"""The app provides a mock snapd API for use with fwupd unit tests. The
/v2/system-secureboot endpoint mimics the behavior of snapd wrt. DBX updates.

The app exposes a couple of test specific endpoinds, /test/scenario, which picks
one of the scenarios (expected to be called in when setting a test fixture),
/test/reset, for resetting the state (usually called in tear down), and
/test/stats which provides the count of actions invoked since last reset.
"""

import argparse
import logging
import base64
import json
import os.path
from io import StringIO
from dataclasses import dataclass
from collections.abc import Callable
from typing import Any, Optional

from flask import Flask, Response, request, current_app


@dataclass
class Scenario:
    handler: Callable[[dict[str, Any]], Response]


@dataclass
class State:
    scenario: Optional[str]
    stats: dict[str, int]

    def __init__(self):
        self.scenario = None
        self.stats = {k: 0 for k in supported_actions}


app = Flask(__name__)

supported_actions = [
    "efi-secureboot-update-startup",
    "efi-secureboot-update-db-cleanup",
    "efi-secureboot-update-db-prepare",
]


def assert_no_extra_keys(d: dict[str, Any], allowed: list[str]):
    unexpected = [a for a in d.keys() if a not in allowed]
    assert len(unexpected) == 0, f"unexpected keys in request data: {unexpected}"


def assert_prepare_req(req: dict[str, Any]):
    assert req.get("action") == "efi-secureboot-update-db-prepare", "unexpected action"
    assert "payload" in req, "missing payload field"
    assert "key-database" in req, "missing key-database field"
    assert req.get("key-database") == "DBX", "unexpected key database"

    payload = req.get("payload")
    assert_no_extra_keys(req, ["action", "payload", "key-database"])
    assert payload is not None, "no update payload"
    # payload must be valid base64
    try:
        raw_payload = base64.urlsafe_b64decode(payload)
        with open(os.path.join(current_app.datadir, "dbx-update.auth"), "rb") as inf:
            reference_payload = inf.read()

        assert raw_payload == reference_payload, "unexpected payload content"
    except Exception as err:
        raise AssertionError("invalid base64 data in request") from err


def assert_startup_req(req: dict[str, Any]):
    assert req.get("action") == "efi-secureboot-update-startup", "unexpected action"
    assert_no_extra_keys(req, ["action"])


def assert_cleanup_req(req: dict[str, Any]):
    assert req.get("action") == "efi-secureboot-update-db-cleanup", "unexpected action"
    assert_no_extra_keys(req, ["action"])


def happy_startup(req: dict[str, Any]) -> Response:
    assert_startup_req(req)
    return Response(None, status=200)


def not_supported(req: dict[str, Any]) -> Response:
    assert_startup_req(req)
    # pretend relevant APIs are missing, hence 404
    return Response("", status=404)


def failed_startup(req: dict[str, Any]) -> Response:
    assert_startup_req(req)
    return Response(None, status=400)


def happy_prepare(req: dict[str, Any]) -> Response:
    assert_prepare_req(req)
    return Response("", status=200)


def failed_prepare(req: dict[str, Any]) -> Response:
    action = req.get("action")
    if action == "efi-secureboot-update-db-cleanup":
        return happy_cleanup(req)
    if action == "efi-secureboot-update-db-prepare":
        assert_prepare_req(req)
        return Response(
            json.dumps(
                {
                    "status": "500",
                    "error": {
                        "kind": "internal-error",
                        "message": "cannot reseal keys in prepare",
                    },
                }
            ),
            status=500,
        )
    if action == "efi-secureboot-update-startup":
        return happy_startup(req)

    raise AssertionError(f"unexpected action {action}")


def failed_cleanup(req: dict[str, Any]) -> Response:
    action = req.get("action")
    if action == "efi-secureboot-update-db-cleanup":
        assert_cleanup_req(req)
        return Response(
            json.dumps(
                {
                    "status": "500",
                    "error": {
                        "kind": "internal-error",
                        "message": "cannot reseal keys in cleanup",
                    },
                }
            ),
            status=500,
        )
    if action == "efi-secureboot-update-db-prepare":
        return happy_prepare(req)
    if action == "efi-secureboot-update-startup":
        return happy_startup(req)

    raise AssertionError(f"unexpected action {action}")


def happy_cleanup(req: dict[str, Any]) -> Response:
    assert_cleanup_req(req)
    return Response("", status=200)


def happy_update(req: dict[str, Any]) -> Response:
    action = req.get("action")
    if action == "efi-secureboot-update-db-cleanup":
        return happy_cleanup(req)
    if action == "efi-secureboot-update-db-prepare":
        return happy_prepare(req)
    if action == "efi-secureboot-update-startup":
        return happy_startup(req)

    raise AssertionError(f"unexpected action {action}")


playbook: dict[str, Scenario] = {
    # successful startup
    "happy-startup": Scenario(handler=happy_startup),
    # startup with mock failure
    "failed-startup": Scenario(handler=failed_startup),
    # 404 on the API endpoint, indicating lack of support on the snapd side
    "not-supported": Scenario(handler=not_supported),
    # prepare step fails
    "failed-prepare": Scenario(handler=failed_prepare),
    # prepare is successful, but cleanup fails
    "failed-cleanup": Scenario(handler=failed_cleanup),
    # successful update cycle
    "happy-update": Scenario(handler=happy_update),
}


def assert_scenario(f):
    def do():
        with app.app_context():
            assert current_app.state.scenario is not None, "test scenario is not set"
            return f()

    return do


def app_init_state():
    with app.app_context():
        current_app.state = State()


@app.route("/v2/system-secureboot", methods=["POST"])
@assert_scenario
def system_secureboot():
    assert len(request.view_args) == 0
    assert request.headers.get("Content-Type") == "application/json"

    req = request.get_json()
    logging.debug("req: %r", req)
    action = req.get("action")

    assert action in supported_actions, f"unknown action {action}"

    current_app.state.stats[action] += 1

    return playbook[current_app.state.scenario].handler(req)


@app.route("/test/setup", methods=["POST"])
def test_setup():
    req = request.get_json()
    logging.debug("req: %r", req)
    scenario = req.get("scenario")

    assert scenario in playbook, f"unknown scenario {scenario}"

    logging.debug("setting scenario to '%s'", scenario)
    current_app.state.scenario = scenario
    return Response("", status=200)


@app.route("/test/reset", methods=["POST"])
def test_reset():
    app_init_state()
    return Response("", status=200)


@app.route("/test/stats")
def test_stats():
    out = StringIO()
    # use a key-file format so that glib side parsing is easy
    out.write("[stats]\n")
    for action in sorted(current_app.state.stats.keys()):
        out.write(f"{action}={current_app.state.stats[action]}\n")
    return Response(out.getvalue(), status=200)


def parse_arguments():
    parser = argparse.ArgumentParser(description="mock snapd APIs")
    parser.add_argument(
        "--socket", help="socket path", default="/tmp/mock-snapd-test.sock"
    )
    parser.add_argument(
        "--datadir",
        default=os.path.dirname(__file__),
        help="path to directory with test data files",
    )

    return parser.parse_args()


if __name__ == "__main__":
    opts = parse_arguments()
    logging.basicConfig(level=logging.DEBUG)
    logging.debug("socket path: %s", opts.socket)
    logging.debug("data dir: %s", opts.datadir)
    with app.app_context():
        current_app.datadir = opts.datadir
    app_init_state()
    app.run(host="unix://" + opts.socket)



#!/usr/bin/env python3
#
# Copyright 2024 Richard Hughes <richard@hughsie.com>
#
# SPDX-License-Identifier: LGPL-2.1-or-later
#
# pylint: disable=invalid-name,missing-docstring

import json
import sys
import hashlib
import zipfile


def _minimize_usb_events(data) -> None:
    for device in data["UsbDevices"]:
        try:
            for event in device.get("UsbEvents", []) + device.get("Events", []):
                if event["Id"].startswith("#"):
                    continue
                event["Id"] = "#" + hashlib.sha1(event["Id"].encode()).hexdigest()[:8]
        except KeyError:
            pass


def _minimize_json(fn: str) -> None:
    with open(fn, "rb") as f:
        data = json.loads(f.read())
        _minimize_usb_events(data)
    with open(fn, "wb") as f:
        f.write(json.dumps(data, indent=2).encode())


def _minimize_zip(fn: str) -> None:
    files = {}
    with zipfile.ZipFile(fn) as myzip:
        for name in myzip.namelist():
            with myzip.open(name) as f:
                data = json.loads(f.read())
                _minimize_usb_events(data)
                files[name] = data
    fn_new = fn.replace(".zip", "-min.zip")
    with zipfile.ZipFile(
        fn_new, "w", compression=zipfile.ZIP_DEFLATED, compresslevel=9
    ) as myzip:
        for name, data in files.items():
            myzip.writestr(name, json.dumps(data, separators=(",", ":")).encode())


if __name__ == "__main__":
    for fn in sys.argv[1:]:
        if fn.endswith(".json"):
            _minimize_json(fn)
        if fn.endswith(".zip"):
            _minimize_zip(fn)



#!/usr/bin/env python3
#
# Copyright 2024 Richard Hughes <richard@hughsie.com>
#
# SPDX-License-Identifier: LGPL-2.1-or-later
#
# pylint: disable=invalid-name,missing-docstring,too-few-public-methods,too-many-locals

import sys
import os
import struct
import subprocess
import io
from typing import List, Optional
import argparse


class IfdPartition:
    REGION_NAMES = [
        "desc",
        "bios",
        "me",
        "gbe",
        "platform",
        "devexp",
        "bios2",
        "ec",
        "ie",
        "10gbe",
    ]

    def __init__(self, region: int = 0, offset: int = 0, size: int = 0) -> None:
        self.region: int = region
        self.offset: int = offset
        self.size: int = size

    def __str__(self) -> str:
        try:
            region_name: str = IfdPartition.REGION_NAMES[self.region]
        except IndexError:
            region_name = "unknown"
        return (
            "IfdPartition("
            f"region=0x{self.region} ({region_name}), "
            f"offset=0x{self.offset:x}, "
            f"size=0x{self.size:x})"
        )


def _read_partitions(f: io.BufferedReader) -> bytearray:
    f.seek(0)
    blob_ifd: bytes = f.read(0x1000)
    (
        signature,
        descriptor_map0,
        descriptor_map1,
        descriptor_map2,
    ) = struct.unpack_from("<16xIIII", blob_ifd, offset=0)
    if signature != 0x0FF0A55A:
        sys.exit(f"Not IFD signature 0x0FF0A55A, got 0x{signature:X}")

    # read out the descriptor maps
    print(f"descriptor_map0=0x{descriptor_map0:X}")
    print(f"descriptor_map1=0x{descriptor_map1:X}")
    print(f"descriptor_map2=0x{descriptor_map2:X}")

    # default to all partitions
    num_regions = (descriptor_map0 >> 24) & 0b111
    if num_regions == 0:
        num_regions = 10
    print(f"num_regions=0x{num_regions:X}")

    # read out FREGs
    flash_region_base_addr = (descriptor_map0 >> 12) & 0x00000FF0
    print(f"flash_region_base_addr=0x{flash_region_base_addr:X}")
    flash_descriptor_regs: List[int] = []
    for region in range(num_regions):
        flash_descriptor_regs.append(
            struct.unpack_from(
                "<I", blob_ifd, offset=flash_region_base_addr + (region * 4)
            )[0]
        )
    for region in range(num_regions):
        print(f"flash_descriptor_reg{region}=0x{flash_descriptor_regs[region]:X}")

    # parse each partition
    fregs: List[IfdPartition] = []
    for i in range(num_regions):
        freg_base: int = (flash_descriptor_regs[i] << 12) & 0x07FFF000
        freg_limit: int = ((flash_descriptor_regs[i] >> 4) & 0x07FFF000) | 0x00000FFF

        # invalid
        if freg_base > freg_limit:
            continue

        fregs.append(
            IfdPartition(region=i, offset=freg_base, size=freg_limit - freg_base)
        )

    # create a binary blob big enough
    image_size: int = 0
    for freg in fregs:
        if freg.offset + freg.size > image_size:
            image_size = freg.offset + freg.size
    print(f"image_size=0x{image_size:x}")
    blob: bytearray = bytearray(image_size)

    # copy each partition
    for freg in fregs:
        print("reading...", freg)
        try:
            f.seek(freg.offset)
            blob_part: bytes = f.read(freg.size)
            blob_size: int = len(blob_part)
            if blob_size != freg.size:
                print(f"tried to read 0x{freg.size:x} and instead got 0x{blob_size:x}")
            if blob_size:
                blob[freg.offset : freg.offset + blob_size] = blob_part
        except OSError as e:
            print(f"failed to read: {e}")
    return blob


def _read_device_to_file(devname: str, filename: Optional[str]) -> None:
    # grab system info from sysfs
    if not filename:
        filename = ""
        for sysfs_fn in [
            "/sys/class/dmi/id/sys_vendor",
            "/sys/class/dmi/id/product_family",
            "/sys/class/dmi/id/product_name",
            "/sys/class/dmi/id/product_sku",
        ]:
            try:
                with open(sysfs_fn, "rb") as f:
                    if filename:
                        filename += "-"
                    filename += (
                        f.read().decode().replace("\n", "").replace(" ", "_").lower()
                    )
            except FileNotFoundError:
                pass
        if filename:
            filename += ".bin"
        else:
            filename = "bios.bin"

    # check this device name is what we expect
    print(f"checking {devname}...")
    try:
        with open(f"/sys/class/mtd/{os.path.basename(devname)}/name", "rb") as f_name:
            name = f_name.read().decode().replace("\n", "")
    except FileNotFoundError as e:
        sys.exit(str(e))
    if name != "BIOS":
        sys.exit(f"Not Intel Corporation PCH SPI Controller, got {name}")

    # read the IFD header, then each partition
    try:
        with open(devname, "rb") as f_in:
            print(f"reading from {devname}...")
            blob = _read_partitions(f_in)
    except PermissionError as e:
        sys.exit(f"cannot read mtd device: {e}")
    print(f"writing {filename}...")
    with open(filename, "wb") as f_out:
        f_out.write(blob)

    # this is really helpful for debugging
    print(f"getting additional data from {devname}...")
    try:
        p = subprocess.run(["mtdinfo", devname], check=True, capture_output=True)
    except subprocess.CalledProcessError as e:
        print(f"{' '.join(args)}: {e}")
    else:
        for line in p.stdout.decode().split("\n"):
            if not line:
                continue
            print(line)
    print("done!")


if __name__ == "__main__":
    # both have defaults
    parser = argparse.ArgumentParser(
        prog="dump-mtd-ifd", description="Dump local SPI contents using MTD"
    )
    parser.add_argument(
        "--filename",
        action="store",
        type=str,
        help="Output filename",
        default=None,
    )
    parser.add_argument(
        "--devname",
        action="store",
        type=str,
        help="Device name, e.g. /dev/mtd0",
        default="/dev/mtd0",
    )
    args = parser.parse_args()
    _read_device_to_file(args.devname, args.filename)



#!/usr/bin/env python3
#
# Copyright 2024 Richard Hughes <richard@hughsie.com>
#
# SPDX-License-Identifier: LGPL-2.1-or-later
#
# pylint: disable=invalid-name,missing-docstring,too-few-public-methods

from typing import Dict
import os
import glob
import sys
import subprocess
import json
import argparse
from collections import defaultdict
from termcolor import colored


def _scan_file(fn: str) -> Dict[str, int]:
    new_map: Dict[str, int] = defaultdict(int)
    try:
        print(f"loading {fn}…")
        args = [
            "./src/fwupdtool",
            "firmware-parse",
            fn,
            "ifd-firmware",
            "--json",
            "--no-timestamp",
        ]
        p = subprocess.run(args, check=True, capture_output=True)
    except subprocess.CalledProcessError as e:
        print(f"{' '.join(args)}: {e}")
    else:
        for line in p.stdout.decode().split("\n"):
            new_map["Lines"] += 1
            if line.find("gtype=") == -1:
                continue
            sections = line.split('"')
            if not sections[1].startswith("Fu"):
                continue
            new_map[sections[1]] += 1
        for line in p.stderr.decode().split("\n"):
            if not line:
                continue
            new_map["WarningLines"] += 1
            print(line)
    return new_map


def _scan_dir(path: str, force_save: bool = False) -> bool:
    all_okay: bool = True
    needs_save: bool = False

    results: Dict[str, Dict[str, int]] = {}

    # support folders or paths
    if os.path.isdir(path):
        for fn in glob.glob(f"{path}/*.bin"):
            results[fn] = _scan_file(fn)
    else:
        results[path] = _scan_file(path)

    # go through each result
    print(f"    {os.path.basename(sys.argv[0])}:")
    for fn, new_map in results.items():
        try:
            with open(f"{fn}.json", "rb") as f:
                old_map = json.loads(f.read().decode())
        except FileNotFoundError:
            old_map = {}
        print(f"    {fn}")
        for key in sorted(set(list(old_map.keys()) + list(new_map.keys()))):
            cnt_old = old_map.get(key, 0)
            cnt_new = new_map.get(key, 0)
            if cnt_new > cnt_old:
                key_str: str = colored(key, "green")
                needs_save = True
            elif cnt_new < cnt_old:
                key_str = colored(key, "red")
                if key != "WarningLines":
                    all_okay = False
            else:
                continue
            print(f"    {key_str:36}: {cnt_old} -> {cnt_new}")

    # save new results if all better
    if (needs_save and all_okay) or force_save:
        for fn, new_map in results.items():
            with open(f"{fn}.json", "wb") as f:
                f.write(json.dumps(new_map, sort_keys=True, indent=4).encode())

    return all_okay


if __name__ == "__main__":
    rc: int = 0

    parser = argparse.ArgumentParser(
        prog="check-ifd-firmware", description="Check IFD firmware parsing"
    )
    parser.add_argument("paths", nargs="+")
    parser.add_argument(
        "--force-save", action="store_true", help="always save the json reports"
    )
    _args = parser.parse_args()
    for _path in _args.paths:
        if not _scan_dir(_path, force_save=_args.force_save):
            rc = 1
    sys.exit(rc)



#!/usr/bin/env python3
#
# Copyright 2024 Mario Limonciello <mario.limonciello@amd.com>
#
# SPDX-License-Identifier: LGPL-2.1-or-later

# to record a string for a device use
# umockdev-record $PATH

import os
import subprocess
import sys
import tempfile
import unittest
import dbusmock

import gi
from gi.repository import GLib
from gi.repository import Gio

gi.require_version("UMockdev", "1.0")
from gi.repository import UMockdev


class FwupdTest(dbusmock.DBusTestCase):
    DBUS_NAME = "org.freedesktop.fwupd"
    DBUS_PATH = "/"
    DBUS_INTERFACE = "org.freedesktop.fwupd"

    @classmethod
    def setUpClass(cls):
        if "DAEMON_BUILDDIR" in os.environ:
            libexecdir = os.environ["DAEMON_BUILDDIR"]
        else:
            libexecdir = "@LIBEXECDIR@"
        cls.daemon_path = os.path.join(libexecdir, "fwupd")
        for dir in ["STATE_DIRECTORY", "CACHE_DIRECTORY"]:
            if dir in os.environ:
                os.makedirs(os.environ[dir], exist_ok=True)
        os.environ["G_DEBUG"] = "fatal_warnings"
        GLib.log_set_always_fatal(
            GLib.LogLevelFlags.LEVEL_WARNING
            | GLib.LogLevelFlags.LEVEL_ERROR
            | GLib.LogLevelFlags.LEVEL_CRITICAL
        )
        # set up a fake system D-BUS
        cls.start_system_bus()
        cls.dbus = Gio.bus_get_sync(Gio.BusType.SYSTEM, None)

    def setUp(self):
        self.testbed = UMockdev.Testbed.new()
        self.polkitd, self.obj_polkit = self.spawn_server_template(
            "polkitd", {}, stdout=subprocess.PIPE
        )
        self.obj_polkit.SetAllowed(
            [
                "org.freedesktop.fwupd.update-internal",
                "org.freedesktop.fwupd.update-internal-trusted",
                "org.freedesktop.fwupd.downgrade-internal-trusted",
                "org.freedesktop.fwupd.downgrade-internal",
                "org.freedesktop.fwupd.update-hotplug-trusted",
                "org.freedesktop.fwupd.update-hotplug",
                "org.freedesktop.fwupd.downgrade-hotplug-trusted",
                "org.freedesktop.fwupd.downgrade-hotplug",
                "org.freedesktop.fwupd.device-unlock",
                "org.freedesktop.fwupd.modify-config",
                "org.freedesktop.fwupd.device-activate",
                "org.freedesktop.fwupd.verify-update",
                "org.freedesktop.fwupd.modify-remote",
                "org.freedesktop.fwupd.set-approved-firmware",
                "org.freedesktop.fwupd.self-sign",
                "org.freedesktop.fwupd.get-bios-settings",
                "org.freedesktop.fwupd.set-bios-settings",
                "org.freedesktop.fwupd.fix-host-security-attr",
                "org.freedesktop.fwupd.undo-host-security-attr",
            ]
        )

        self.proxy = None
        self.props_proxy = None
        self.daemon_log = None
        self.daemon = None
        self.changed_properties = {}

    def run(self, result=None):
        super().run(result)
        if not result or not self.daemon_log:
            return
        if len(result.errors) + len(result.failures) or os.getenv("VERBOSE"):
            with open(self.daemon_log.name, encoding="utf-8") as tmpf:
                sys.stderr.write("\nDAEMON log:\n")
                sys.stderr.write(tmpf.read())
                sys.stderr.write("\n")

    def tearDown(self):
        del self.testbed
        self.stop_daemon()

        if self.polkitd:
            self.polkitd.stdout.close()
            try:
                self.polkitd.kill()
            except OSError:
                pass
            self.polkitd.wait()

        self.obj_polkit = None

    #
    # Daemon control and D-BUS I/O
    #

    def start_daemon(self):
        """Start daemon and create DBus proxy.

        When done, this sets self.proxy as the Gio.DBusProxy for power-profiles-daemon.
        """
        env = os.environ.copy()
        env["G_DEBUG"] = "fatal-criticals"
        env["G_MESSAGES_DEBUG"] = "all"
        # note: Python doesn't propagate the setenv from Testbed.new(), so we
        # have to do that ourselves
        env["UMOCKDEV_DIR"] = self.testbed.get_root_dir()
        self.daemon_log = (
            tempfile.NamedTemporaryFile()
        )  # pylint: disable=consider-using-with
        daemon_path = [self.daemon_path, "-vv"]

        # pylint: disable=consider-using-with
        self.daemon = subprocess.Popen(
            daemon_path, env=env, stdout=self.daemon_log, stderr=subprocess.STDOUT
        )
        self.addCleanup(self.daemon.kill)

        def on_proxy_connected(_, res):
            try:
                self.proxy = Gio.DBusProxy.new_finish(res)
            except GLib.Error as exc:
                self.fail(exc)

        cancellable = Gio.Cancellable()
        self.addCleanup(cancellable.cancel)
        Gio.DBusProxy.new(
            self.dbus,
            Gio.DBusProxyFlags.DO_NOT_AUTO_START,
            None,
            self.DBUS_NAME,
            self.DBUS_PATH,
            self.DBUS_INTERFACE,
            cancellable,
            on_proxy_connected,
        )

        # wait until the daemon gets online
        wait_time = 60 if "valgrind" in daemon_path[0] else 5
        self.assert_eventually(
            lambda: self.proxy and self.proxy.get_name_owner(),
            timeout=wait_time * 1000,
            message=f"daemon did not start in {wait_time} seconds",
        )

        def properties_changed_cb(_, changed_properties, invalidated):
            self.changed_properties.update(changed_properties.unpack())

        self.addCleanup(
            self.proxy.disconnect,
            self.proxy.connect("g-properties-changed", properties_changed_cb),
        )

        self.assertEqual(self.daemon.poll(), None, "daemon crashed")

    def stop_daemon(self):
        """Stop the daemon if it is running."""

        if self.daemon:
            try:
                self.daemon.terminate()
            except OSError:
                pass
            self.assertEqual(self.daemon.wait(timeout=3000), 0)

        self.daemon = None
        self.proxy = None

    def assert_eventually(self, condition, message=None, timeout=5000):
        """Assert that condition function eventually returns True.

        Timeout is in milliseconds, defaulting to 5000 (5 seconds). message is
        printed on failure.
        """
        if condition():
            return

        done = False

        def on_timeout_reached():
            nonlocal done
            done = True

        source = GLib.timeout_add(timeout, on_timeout_reached)
        while not done:
            if condition():
                GLib.source_remove(source)
                return
            GLib.MainContext.default().iteration(False)

        self.fail(message or "timed out waiting for " + str(condition))

    def assert_dbus_property_eventually_is(self, prop, value, timeout=1200):
        """Asserts that a dbus property eventually is what expected"""
        return self.assert_eventually(
            lambda: self.get_dbus_property(prop) == value,
            timeout=timeout,
            message=f"property '{prop}' is not '{value}', but "
            + f"'{self.get_dbus_property(prop)}'",
        )

    def ensure_dbus_properties_proxies(self):
        self.props_proxy = Gio.DBusProxy.new_sync(
            self.dbus,
            Gio.DBusProxyFlags.DO_NOT_AUTO_START
            | Gio.DBusProxyFlags.DO_NOT_AUTO_START_AT_CONSTRUCTION
            | Gio.DBusProxyFlags.DO_NOT_LOAD_PROPERTIES
            | Gio.DBusProxyFlags.DO_NOT_CONNECT_SIGNALS,
            None,
            self.DBUS_NAME,
            self.DBUS_PATH,
            "org.freedesktop.DBus.Properties",
            None,
        )

    def get_dbus_property(self, name):
        """Get property value from daemon D-Bus interface."""
        self.ensure_dbus_properties_proxies()
        return self.props_proxy.Get("(ss)", self.DBUS_NAME, name)

    def start_upower(self, on_battery):
        """Start the upower service with the given on_battery state"""

        sys_bat = self.testbed.add_device(
            "power_supply",
            "fakeBAT0",
            None,
            [
                "type",
                "Battery",
                "present",
                "1",
                "status",
                "Discharging",
                "energy_full",
                "60000000",
                "energy_full_design",
                "80000000",
                "energy_now",
                "48000000",
                "voltage_now",
                "12000000",
            ],
            ["POWER_SUPPLY_ONLINE", "1"],
        )

        self.upowerd, obj_upower = self.spawn_server_template(
            "upower",
            {"DaemonVersion": "0.99", "OnBattery": on_battery},
            stdout=subprocess.PIPE,
        )

    def stop_upower(self):
        """Stop the upower service"""
        self.upowerd.terminate()
        self.upowerd.wait()


class VeryBasicTest(FwupdTest):
    """Test the basic properties of the daemon."""

    @classmethod
    def setUpClass(cls):
        super().setUpClass()

        gi.require_version("Fwupd", "2.0")
        from gi.repository import Fwupd  # pylint: disable=wrong-import-position

        cls.client = Fwupd.Client()

    def test_properties(self):
        """Test the properties of the daemon without a client."""
        self.start_daemon()

        self.assertEqual(self.get_dbus_property("Interactive"), False)
        self.assertEqual(self.get_dbus_property("OnlyTrusted"), True)
        self.assertEqual(self.get_dbus_property("Tainted"), False)
        self.assertEqual(self.get_dbus_property("HostBkc"), "")
        self.assertEqual(self.get_dbus_property("HostProduct"), "Unknown Product")
        self.assertEqual(self.get_dbus_property("HostVendor"), "Unknown Vendor")
        self.assertEqual(self.get_dbus_property("Percentage"), 0)
        self.assertEqual(self.get_dbus_property("Status"), 1)
        self.assertEqual(self.get_dbus_property("BatteryLevel"), 101)

    def test_get_devices(self):
        """Test the libfwupd client get-devices interface."""
        self.start_daemon()

        devices = self.client.get_devices()
        # Should be at least the CPU test is running on
        self.assertGreater(len(devices), 0)


if __name__ == "__main__":
    # run ourselves under umockdev
    if "umockdev" not in os.environ.get("LD_PRELOAD", ""):
        os.execvp("umockdev-wrapper", ["umockdev-wrapper", sys.executable] + sys.argv)

    prog = unittest.main(exit=False)
    if prog.result.errors or prog.result.failures:
        sys.exit(1)

    # Translate to skip error
    if prog.result.testsRun == len(prog.result.skipped):
        sys.exit(77)
